from typing import Any, TypeVar

T_in = TypeVar("T_in", default=Any, infer_variance=True)
"""Input data type for pipeline flow (payload's value type)"""

T_out = TypeVar("T_out", default=T_in, infer_variance=True)
"""Output data type for pipeline flow (result's value type)"""



import pytest

from thecodecrate_pipeline import PipelineFactory, Stage
from thecodecrate_pipeline.processors import ChainedProcessor
from thecodecrate_pipeline.types import StageDefinitionCollection

from .stubs.stub_stages_int import (
    AddOneStage,
)

some_stages: StageDefinitionCollection = (
    AddOneStage,
    (lambda x: x + 1),
    (lambda x: f"result is {x}"),
)


@pytest.mark.asyncio
async def test_with_stages():
    pipeline_factory = (PipelineFactory[int, str]()).with_stages(some_stages)

    pipeline = pipeline_factory.make()

    assert await pipeline.process(10) == "result is 12"


@pytest.mark.asyncio
async def test_parameter_immutability():
    # creating a pipeline with a few stages
    pipeline_factory = (PipelineFactory()).with_stages(some_stages)
    pipeline = pipeline_factory.make()
    result = await pipeline.process(10)
    assert result == "result is 12"

    # let's try to change the stages list
    def add_excitement(x: str) -> str:
        return f"{x}!!!"

    pipeline_factory.add_stage(add_excitement)

    # the original "some_stages" list should not be affected
    result2 = await pipeline.process(10)
    assert result2 == result
    assert len(some_stages) == 3


@pytest.mark.asyncio
async def test_immutability():
    # creating a pipeline with a few stages
    pipeline_factory = (PipelineFactory[int, str]()).with_stages(some_stages)
    pipeline = pipeline_factory.make()
    result = await pipeline.process(10)
    assert result == "result is 12"

    # adding another stage after the pipeline has been created
    def add_excitement(x: str) -> str:
        return f"{x}!!!"

    pipeline_factory.add_stage(add_excitement)

    # the original pipeline is not affected
    result2 = await pipeline.process(10)
    assert result2 == result

    # now, let's create a new pipeline, with the excitement stage!!!
    pipeline2 = pipeline_factory.make()
    result3 = await pipeline2.process(10)

    # the new pipeline has the excitement!!!
    assert result3 == "result is 12!!!"

    # notice the original pipeline is not affected
    result4 = await pipeline.process(10)
    assert result4 == result

    # what if we try to change the original pipeline, directly?
    pipeline.pipe(
        add_excitement
    )  # <- adding the excitement stage to the original pipeline

    # the original pipeline is still not affected, due to the immutability
    # nature of pipelines
    result5 = await pipeline.process(10)
    assert result5 == result


@pytest.mark.asyncio
async def test_with_processor():
    class MyProcessor(ChainedProcessor):
        pass

    # create factory with a processor
    pipeline_factory = (
        (PipelineFactory[int, str]())
        .with_stages(some_stages)
        .with_processor(MyProcessor())
    )

    # create and process
    pipeline = pipeline_factory.make()
    result = await pipeline.process(10)
    assert result == "result is 12"

    # check the processor
    assert pipeline.get_processor_instance().__class__ == MyProcessor


@pytest.mark.asyncio
async def test_with_class_stages():
    class MyStage(Stage[int, str]):
        async def __call__(self, x: int) -> str:
            return f"result is {x}"

    pipeline_factory = (PipelineFactory[int, str]()).with_stages((MyStage,))

    pipeline = pipeline_factory.make()
    assert await pipeline.process(10) == "result is 10"


@pytest.mark.asyncio
async def test_with_declared_stages():
    class MyStage(Stage[int, str]):
        async def __call__(self, x: int) -> str:
            return f"result is {x}"

    class MyPipelineFactory(PipelineFactory[int, str]):
        stages = (MyStage,)

    pipeline_factory = MyPipelineFactory()

    pipeline = pipeline_factory.make()
    assert await pipeline.process(10) == "result is 10"



import inspect
from typing import Any, Awaitable, Callable, cast

from ...processor import Processor
from ...types.callable_type import CallableCollection
from ...types.types import T_in, T_out

# Type for a callable that checks a condition on the payload.
CheckCallable = Callable[[T_in], bool | Awaitable[bool]]


class InterruptibleProcessor(Processor[T_in, T_out]):
    """Processor with conditional interruption."""

    check: CheckCallable[T_in]
    """Callable for processing interruption.

    Example:
        ```python
        class MaxValueProcessor(InterruptibleProcessor[int, int]):
            # interrupt if value exceeds 100
            check = lambda x: x > 100
        ```
    """

    def __init__(self, check: CheckCallable[T_in]) -> None:
        """Constructor.

        Parameters:
            check: Callable for processing interruption.

        Example:
            ```python
            # Interrupts when payload value exceeds 100
            def check_value(payload: int) -> bool:
                return payload > 100

            # Create processor with the check
            processor = InterruptibleProcessor(check_value)

            # Process payload - will stop if value exceeds 100
            result = await processor.process(initial_payload, stages)
            ```
        """
        super().__init__()

        self.check = check

    async def process(
        self,
        payload: T_in,
        stages: CallableCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        payload_out: Any = payload

        for stage in stages:
            payload_out = await self._call(
                callable=stage, payload=payload_out, *args, **kwds
            )

            if await self._call_check(payload_out):
                return cast(T_out, payload_out)

        return cast(T_out, payload_out)

    async def _call_check(self, payload: T_in) -> bool:
        result = self.check(payload)

        if inspect.isawaitable(result):
            return await result

        return result



import copy
from typing import Any, Protocol, Self

from .clonable_interface import ClonableInterface


class Clonable(ClonableInterface, Protocol):
    def clone(self, attributes: dict[str, Any]) -> Self:
        cloned = copy.deepcopy(self)

        for key, value in attributes.items():
            setattr(cloned, key, value)

        return cloned



"""A collection of processors and their pipelines"""

# Re-exporting symbols
from _lib.processors.chained_processor.chained_pipeline import (
    ChainedPipeline as ChainedPipeline,
)
from _lib.processors.chained_processor.chained_processor import (
    ChainedProcessor as ChainedProcessor,
)
from _lib.processors.interruptible_processor.interruptible_pipeline import (
    InterruptiblePipeline as InterruptiblePipeline,
)
from _lib.processors.interruptible_processor.interruptible_processor import (
    InterruptibleProcessor as InterruptibleProcessor,
)

__all__ = (
    "ChainedPipeline",
    "ChainedProcessor",
    "InterruptiblePipeline",
    "InterruptibleProcessor",
)



from abc import ABC
from typing import Any, Optional, Self

from ..contracts.processable_pipeline import ProcessablePipeline as ImplementsInterface
from ..contracts.processor import Processor as ProcessorContract
from ..processors.chained_processor.chained_processor import ChainedProcessor
from ..types.types import T_in, T_out


class ProcessablePipeline(
    ImplementsInterface[T_in, T_out],
    ABC,
):
    processor: Optional[type[ProcessorContract] | ProcessorContract]
    processor_instance: Optional[ProcessorContract]

    def __init__(
        self,
        processor: Optional[type[ProcessorContract] | ProcessorContract] = None,
        processor_instance: Optional[ProcessorContract] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if not hasattr(self, "processor"):
            self.processor = self._get_default_processor()

        if not hasattr(self, "processor_instance"):
            self.processor_instance = None

        if processor:
            self.processor = processor

        if processor_instance:
            self.processor_instance = processor_instance

        if self._should_instantiate_processor():
            self._instantiate_processor()

    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out:
        """
        Process the given payload through the pipeline.
        """
        if self.processor_instance is None:
            raise ValueError("Processor not set")

        return await self.processor_instance.process(
            payload=payload, stages=self.get_stages_instances(), *args, **kwds
        )

    async def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Processes payload through the pipeline.
        """
        return await self.process(payload, *args, **kwds)

    def with_processor(
        self, processor: type[ProcessorContract] | ProcessorContract
    ) -> Self:
        """
        Attachs a processor (class or instance) to the pipeline.
        """
        cloned = self.clone({"processor": processor, "processor_instance": None})

        return cloned._instantiate_processor()

    def get_processor_instance(self) -> Optional[ProcessorContract]:
        return self.processor_instance

    def _get_default_processor(self) -> type[ChainedProcessor[T_in, T_out]]:
        return ChainedProcessor

    def _should_instantiate_processor(self) -> bool:
        return self.processor_instance is None

    def _instantiate_processor(self) -> Self:
        if self.processor is None:
            raise ValueError("Processor class not set")

        if isinstance(self.processor, type):
            self.processor_instance = self.processor()
        else:
            self.processor_instance = self.processor

        if isinstance(self.processor_instance, type):
            raise ValueError("Processor instance could not be created")

        return self



from typing import Protocol

from ..support.clonable import ClonableInterface
from ..types.types import T_in, T_out


class BasePipeline(
    ClonableInterface,
    Protocol[T_in, T_out],
):
    pass



from typing import TypeVar

from .concerns.base_pipeline import BasePipeline
from .concerns.processable_pipeline import ProcessablePipeline
from .concerns.stageable_pipeline import StageablePipeline
from .contracts.pipeline import Pipeline as ImplementsInterface
from .types.types import T_in, T_out


class Pipeline(
    ProcessablePipeline[T_in, T_out],
    StageablePipeline[T_in, T_out],
    BasePipeline[T_in, T_out],
    ImplementsInterface[T_in, T_out],
):
    pass


TPipeline = TypeVar("TPipeline", bound=Pipeline, infer_variance=True)



from abc import abstractmethod
from typing import Any, Protocol

from ..types.types import T_in, T_out


class Stage(
    Protocol[T_in, T_out],
):
    @abstractmethod
    async def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...



from typing import Any, cast

from ...processor import Processor
from ...types.callable_type import CallableCollection
from ...types.types import T_in, T_out


class ChainedProcessor(
    Processor[T_in, T_out],
):
    async def process(
        self,
        payload: T_in,
        stages: CallableCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Process the given payload through the provided stages.

        Args:
            payload (T_in): The input payload to process.
            stages (CallableCollection): The collection of stages to process the payload through.
            *args (Any): Additional positional arguments.
            **kwds (Any): Additional keyword arguments.

        Returns:
            T_out: The processed output.
        """
        payload_out: Any = payload

        for stage in stages:
            payload_out = await self._call(
                callable=stage, payload=payload_out, *args, **kwds
            )

        return cast(T_out, payload_out)



from typing import Protocol

from ..types.types import T_in, T_out
from .base_pipeline import BasePipeline as BasePipelineContract
from .processable_pipeline import ProcessablePipeline as ProcessablePipelineContract
from .stageable_pipeline import StageablePipeline as StageablePipelineContract


class Pipeline(
    ProcessablePipelineContract[T_in, T_out],
    StageablePipelineContract[T_in, T_out],
    BasePipelineContract[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass



from abc import abstractmethod
from typing import Any, Awaitable, Protocol

from ..support.clonable.clonable_interface import ClonableInterface
from ..types.callable_type import CallableCollection, CallableType
from ..types.types import T_in, T_out


class Processor(
    ClonableInterface,
    Protocol[T_in, T_out],
):
    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: CallableCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...

    async def _call(
        self,
        callable: CallableType[T_in, T_out | Awaitable[T_out]],
        payload: T_in,
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...



from abc import abstractmethod
from typing import Any, Optional

from .act_as_factory_interface import ActAsFactoryInterface, T_target


class ActAsFactory(ActAsFactoryInterface[T_target]):
    # set the class here or in `_get_target_class`
    _target_class: Optional[type[T_target]] = None

    @abstractmethod
    def _definition(self) -> dict[str, Any]:
        pass

    def _get_target_class(self) -> type[T_target]:
        if self._target_class is None:
            raise ValueError("Factory _target_class not set")

        return self._target_class

    def make(self) -> T_target:
        instance_class = self._get_target_class()

        return instance_class(**self._definition())



from typing import Any, Awaitable, Protocol

from .types import T_in, T_out


class CallableType(
    Protocol[T_in, T_out],
):
    """Protocol for callable stages that process data in a pipeline."""

    def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out | Awaitable[T_out]: ...


CallableCollection = tuple[CallableType, ...]
"""Collection of objects or functions used as stages in the pipeline"""

StageDefinition = CallableType | type[CallableType]
"""A Stage class, object or function"""

StageDefinitionCollection = tuple[StageDefinition, ...]
"""Collection of Stage classes, objects or functions used to define stages in the pipeline"""



from typing import Any

from ...pipeline import Pipeline, T_in, T_out
from .chained_processor import ChainedProcessor


class ChainedPipeline(Pipeline[T_in, T_out]):
    """Default pipeline (`Pipeline` alias). Sequentially processes data through multiple stages."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Constructor.

        Example:
            ```python
            # Process data through multiple stages
            pipeline = (
                (ChainedPipeline[int]())
                .pipe(lambda payload: payload + 1)
                .pipe(lambda payload: payload * 2)
                .pipe(lambda payload: payload + 1)
            )

            # Assert result
            assert await pipeline.process(1) == 5
            ```
        """
        super().__init__(
            processor=ChainedProcessor[T_in, T_out](),
            *args,
            **kwargs,
        )



import inspect
from abc import abstractmethod
from typing import Any

from .contracts.processor import Processor as ImplementsInterface
from .support.clonable import Clonable
from .types.callable_type import CallableCollection, CallableType
from .types.types import T_in, T_out


class Processor(
    Clonable,
    ImplementsInterface[T_in, T_out],
):
    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: CallableCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Process the given payload through the provided stages.

        Args:
            payload (T_in): The input payload to process.
            stages (CallableCollection): The collection of stages to process the payload through.
            *args (Any): Additional positional arguments.
            **kwds (Any): Additional keyword arguments.

        Returns:
            T_out: The processed output.
        """
        pass

    async def _call(
        self,
        callable: CallableType[T_in, T_out],
        payload: T_in,
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Do the actual processing of the payload - the "process" method is just an alias to this method.
        """
        result = callable(payload, *args, **kwds)

        if inspect.isawaitable(result):
            return await result

        return result



# Version of the package
# DO NOT MODIFY MANUALLY
# This will be updated by `bumpver` command.
# - Make sure to commit all changes first before running `bumpver`.
# - Run `bumpver update --[minor|major|patch]`
__version__ = "1.28.0"

# Re-exporting symbols
from _lib.contracts.pipeline import Pipeline as PipelineInterface
from _lib.contracts.pipeline_factory import PipelineFactory as PipelineFactoryInterface
from _lib.contracts.processor import Processor as ProcessorInterface
from _lib.contracts.stage import Stage as StageInterface
from _lib.pipeline import Pipeline as Pipeline
from _lib.pipeline_factory import PipelineFactory as PipelineFactory
from _lib.processor import Processor as Processor
from _lib.stage import Stage as Stage

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    "Pipeline",
    "PipelineFactory",
    "PipelineFactoryInterface",
    "PipelineInterface",
    "Processor",
    "ProcessorInterface",
    "Stage",
    "StageInterface",
)



from typing import Any, Optional, Protocol, Self

from ..types.callable_type import (
    CallableCollection,
    CallableType,
    StageDefinitionCollection,
)
from ..types.types import T_in, T_out
from .base_pipeline import BasePipeline as BasePipelineContract


class StageablePipeline(
    BasePipelineContract[T_in, T_out],
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        stages: Optional[StageDefinitionCollection] = None,
        stages_instances: Optional[CallableCollection] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def pipe(self, stage: CallableType) -> Self: ...

    def with_stages(self, stages: StageDefinitionCollection) -> Self: ...

    def get_stages(self) -> StageDefinitionCollection: ...

    def get_stages_instances(self) -> CallableCollection: ...



from abc import abstractmethod
from typing import Any, Awaitable, Callable, Concatenate, cast

from thecodecrate_pipeline import Pipeline, Processor, Stage
from thecodecrate_pipeline.types import CallableCollection, T_in, T_out


class IndexedStage(Stage[T_in, T_out]):
    @abstractmethod
    async def __call__(
        self,
        payload: T_in,
        /,
        tag: int,
    ) -> T_out:
        pass


IndexedPipelineCallable = (
    IndexedStage[T_in, T_out]
    | Callable[Concatenate[T_in, ...], Awaitable[T_out]]
    | Callable[Concatenate[T_in, ...], T_out]
)


class IndexedProcessor(Processor[T_in, T_out]):
    async def process(
        self,
        payload: T_in,
        stages: CallableCollection,
    ) -> T_out:
        index = 0

        payload_out: Any = payload

        for stage in stages:
            payload_out = await self._call(
                callable=stage, payload=payload_out, index=index
            )

            index += 1

        return cast(T_out, payload_out)


class IndexedPipeline(Pipeline[T_in]):
    processor = IndexedProcessor



from thecodecrate_pipeline import Stage


class TimesTwoStage(Stage[int]):
    async def __call__(self, payload: int) -> int:
        return payload * 2


class AddOneStage(Stage[int]):
    async def __call__(self, payload: int) -> int:
        return payload + 1


class InvalidStage(Stage[int]):
    async def __call__(self, payload: str) -> int:
        return len(payload)


class TimesThreeStage(Stage[int]):
    async def __call__(self, payload: int) -> int:
        return payload * 3



from abc import ABC
from typing import Any, Optional, Self

from ..contracts.stageable_pipeline import StageablePipeline as ImplementsInterface
from ..types.callable_type import (
    CallableCollection,
    CallableType,
    StageDefinitionCollection,
)
from ..types.types import T_in, T_out


class StageablePipeline(
    ImplementsInterface[T_in, T_out],
    ABC,
):
    stages: StageDefinitionCollection
    stages_instances: CallableCollection

    def __init__(
        self,
        stages: Optional[StageDefinitionCollection] = None,
        stages_instances: Optional[CallableCollection] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if not hasattr(self, "stages"):
            self.stages = tuple()

        if not hasattr(self, "stages_instances"):
            self.stages_instances = tuple()

        if stages:
            self.stages = stages

        if stages_instances:
            self.stages_instances = stages_instances

        if self._should_instantiate_stages():
            self._instantiate_stages()

    def pipe(self, stage: CallableType) -> Self:
        """
        Adds a single stage to the pipeline.
        """
        return self.clone({"stages_instances": tuple([*self.stages_instances, stage])})

    def with_stages(self, stages: StageDefinitionCollection) -> Self:
        """
        Adds a collection of stages to the pipeline.
        """
        cloned = self.clone({"stages": stages, "stages_instances": []})

        return cloned._instantiate_stages()

    def get_stages(self) -> StageDefinitionCollection:
        return self.stages

    def get_stages_instances(self) -> CallableCollection:
        return self.stages_instances

    def _should_instantiate_stages(self) -> bool:
        return len(self.stages_instances) == 0 and len(self.stages) > 0

    def _instantiate_stages(self) -> Self:
        self.stages_instances = tuple(
            stage() if isinstance(stage, type) else stage for stage in self.stages
        )

        return self



from typing import Any, Optional, Protocol, Self

from ..types.types import T_in, T_out
from .base_pipeline import BasePipeline as BasePipelineContract
from .processor import Processor as ProcessorContract
from .stageable_pipeline import StageablePipeline as StageablePipelineContract


class ProcessablePipeline(
    StageablePipelineContract[T_in, T_out],
    BasePipelineContract[T_in, T_out],
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        processor: Optional[type[ProcessorContract] | ProcessorContract] = None,
        processor_instance: Optional[ProcessorContract] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out: ...

    async def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...

    def with_processor(
        self, processor: type[ProcessorContract] | ProcessorContract
    ) -> Self: ...

    def get_processor_instance(self) -> Optional[ProcessorContract]: ...



from thecodecrate_pipeline import StageInterface


class StubStage(StageInterface[str]):
    STUBBED_RESPONSE = "stubbed response"

    async def __call__(self, payload: str) -> str:
        return self.STUBBED_RESPONSE



from typing import Any, Optional, Self

from .contracts.pipeline import Pipeline as PipelineContract
from .contracts.pipeline_factory import PipelineFactory as ImplementsInterface
from .contracts.processor import Processor as ProcessorContract
from .pipeline import Pipeline
from .support.act_as_factory.act_as_factory import ActAsFactory
from .types.callable_type import StageDefinition, StageDefinitionCollection
from .types.types import T_in, T_out


class PipelineFactory(
    ActAsFactory[PipelineContract[T_in, T_out]],
    ImplementsInterface[T_in, T_out],
):
    stages: StageDefinitionCollection
    processor: Optional[type[ProcessorContract[T_in, T_out]] | ProcessorContract]
    pipeline_class: Optional[type[PipelineContract[T_in, T_out]]]

    def __init__(
        self,
        stages: Optional[StageDefinitionCollection] = None,
        processor: Optional[type[ProcessorContract] | ProcessorContract] = None,
        pipeline_class: Optional[type[PipelineContract[T_in, T_out]]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        if not hasattr(self, "stages"):
            self.stages = tuple()

        if not hasattr(self, "processor"):
            self.processor = None

        if not hasattr(self, "pipeline_class"):
            self.pipeline_class = self._get_default_pipeline_class()

        if stages:
            self.stages = stages

        if processor:
            self.with_processor(processor)

        if pipeline_class:
            self.pipeline_class = pipeline_class

    def add_stage(self, stage: StageDefinition) -> Self:
        """
        Adds a single stage to the pipeline.
        """
        self.stages = self.stages + (stage,)

        return self

    def with_stages(self, stages: StageDefinitionCollection) -> Self:
        """
        Adds a collection of stages to the pipeline.
        """
        self.stages = stages

        return self

    def with_processor(
        self, processor: type[ProcessorContract] | ProcessorContract
    ) -> Self:
        """
        Attachs a processor (class or instance) to the pipeline factory.
        """
        self.processor = processor

        return self

    def _get_default_pipeline_class(
        self,
    ) -> Optional[type[PipelineContract[T_in, T_out]]]:
        return Pipeline

    # ActAsFactory
    def _definition(self) -> dict[str, Any]:
        return {
            "stages": self.stages,
            "processor": self.processor,
        }

    # ActAsFactory
    def _get_target_class(self) -> type[PipelineContract[T_in, T_out]]:
        if self.pipeline_class is None:
            raise ValueError("Pipeline class not set in factory.")

        return self.pipeline_class



from abc import ABC
from typing import Any, Generic

from ..support.clonable import Clonable
from ..types.types import T_in, T_out


class BasePipeline(
    Clonable,
    Generic[T_in, T_out],
    ABC,
):
    def __init__(
        self,
        *args: Any,
        **kwds: Any,
    ) -> None:
        pass



import pytest

from thecodecrate_pipeline import Pipeline
from thecodecrate_pipeline.types import CallableCollection, StageDefinitionCollection

from .stubs.stub_processor import StubProcessor
from .stubs.stub_stage import StubStage
from .stubs.stub_stage_with_custom_args import (
    IndexedPipeline,
    IndexedProcessor,
    IndexedStage,
)
from .stubs.stub_stages_int import (
    AddOneStage,
    TimesThreeStage,
    TimesTwoStage,
)


@pytest.mark.asyncio
async def test_lambda_stages():
    pipeline = (
        (Pipeline[int]())
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
        .pipe(lambda payload: payload + 1)
    )

    assert await pipeline.process(1) == 5
    assert await pipeline.process(2) == 7


@pytest.mark.asyncio
async def test_classbased_stages():
    pipeline = (
        (Pipeline[int]())
        .pipe(TimesTwoStage())
        .pipe(AddOneStage())
        .pipe(TimesThreeStage())
    )

    assert await pipeline.process(5) == 33


@pytest.mark.asyncio
async def test_pipelinebased_stages():
    sub_pipeline1 = (
        (Pipeline[int]())
        .pipe(TimesTwoStage())
        .pipe(AddOneStage())
        .pipe(TimesThreeStage())
    )

    sub_pipeline2 = (
        (Pipeline[int]())
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
    )

    pipeline = (
        (Pipeline[int]())
        .pipe(TimesThreeStage())
        .pipe(sub_pipeline1)
        .pipe(sub_pipeline2)
        .pipe(lambda payload: payload * 2)
    )

    assert await pipeline.process(10) == 736


@pytest.mark.asyncio
async def test_string_stage():
    pipeline = (Pipeline[str]()).pipe(StubStage())

    assert await pipeline.process("") == "stubbed response"


@pytest.mark.asyncio
async def test_stages_with_different_types():
    pipeline = (
        (Pipeline[int, str]())
        .pipe(lambda x: x + 1)
        .pipe(lambda payload: f"total is {payload}!")
    )

    assert await pipeline.process(10) == "total is 11!"


@pytest.mark.asyncio
async def test_process_repeat():
    # it happened before that the pipeline was not reset,
    # making the second process call would not work as expected.
    for _ in range(3):
        pipeline = (
            (Pipeline[int]())
            .pipe(lambda payload: payload + 1)
            .pipe(lambda payload: payload * 2)
            .pipe(lambda payload: payload + 1)
        )

        for __ in range(3):
            assert await pipeline.process(1) == 5
            assert await pipeline.process(2) == 7


@pytest.mark.asyncio
async def test_pipeline_immutability():
    pipeline = (
        (Pipeline[int]())
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload + 1)
    )

    assert await pipeline.process(1) == 3

    pipeline2 = pipeline.pipe(lambda payload: payload + 1)

    assert await pipeline.process(1) == 3
    assert await pipeline2.process(1) == 4


@pytest.mark.asyncio
async def test_declarative_stages():
    pipeline = (Pipeline[int]()).pipe(TimesTwoStage()).pipe(AddOneStage())

    def add_seven(payload: int) -> int:
        return payload + 7

    async def sub_three_async(payload: int) -> int:
        return payload - 3

    class MyPipeline(Pipeline[int]):
        stages = (
            TimesThreeStage(),  # stage instance
            pipeline,  # pipeline
            AddOneStage,  # stage class
            add_seven,  # function
            sub_three_async,  # async function
        )

    assert await MyPipeline().process(5) == 36


@pytest.mark.asyncio
async def test_declarative_stages_with_processor():
    class MyPipeline(Pipeline[int]):
        processor = StubProcessor
        stages = (
            TimesTwoStage,
            TimesThreeStage,
        )

    pipeline = MyPipeline()

    assert await pipeline.process(5) == 300

    assert pipeline.__class__ == MyPipeline
    assert pipeline.processor_instance.__class__ == StubProcessor


@pytest.mark.asyncio
async def test_declarative_stages_instances():
    pipeline = (Pipeline[int]()).pipe(TimesTwoStage()).pipe(AddOneStage())

    def add_seven(payload: int) -> int:
        return payload + 7

    async def sub_three_async(payload: int) -> int:
        return payload - 3

    class MyPipeline(Pipeline[int]):
        stages_instances = (
            TimesThreeStage(),  # stage instance
            pipeline,  # pipeline
            AddOneStage(),  # another stage instance
            add_seven,  # function
            sub_three_async,  # async function
        )

    stages_instances: CallableCollection = (
        add_seven,  # function
        sub_three_async,  # async function
    )

    assert await MyPipeline().process(5) == 36
    assert await MyPipeline(stages_instances=stages_instances).process(5) == 9


@pytest.mark.asyncio
async def test_declarative_stages_instances__immutability():
    def add_seven(payload: int) -> int:
        return payload + 7

    async def sub_three_async(payload: int) -> int:
        return payload - 3

    stages_instances: CallableCollection = (
        add_seven,  # function
        sub_three_async,  # async function
    )

    pipeline = Pipeline(stages_instances=stages_instances)

    assert await pipeline.process(5) == 9

    # adding more stages should not affect the original pipeline
    stages_instances += (TimesTwoStage(),)

    assert await pipeline.process(5) == 9


@pytest.mark.asyncio
async def test_constructor_override_declared_stages():
    class MyPipeline(Pipeline[int]):
        stages = (
            TimesTwoStage,  # stage class
            TimesThreeStage(),  # stage instance
        )

    stages: StageDefinitionCollection = (
        AddOneStage,  # stage class
        TimesTwoStage(),  # stage instance
    )

    stages_instances: CallableCollection = (
        AddOneStage(),  # only instances can be added
        TimesTwoStage(),  # only instances can be added
    )

    assert await MyPipeline().process(5) == 30
    assert await MyPipeline(stages=stages).process(5) == 12
    assert await MyPipeline(stages=stages_instances).process(5) == 12
    assert await MyPipeline(stages_instances=stages_instances).process(5) == 12


@pytest.mark.asyncio
async def test_custom_args_on_stages():
    class MyIndexedStage(IndexedStage[str]):
        async def __call__(self, payload: str, index: int) -> str:
            return f"{payload}: {index}"

    pipeline = (IndexedPipeline[str]()).pipe(MyIndexedStage()).pipe(MyIndexedStage())

    assert await pipeline.process("test") == "test: 0: 1"
    assert pipeline.processor_instance.__class__ == IndexedProcessor
    assert pipeline.__class__ == IndexedPipeline


@pytest.mark.asyncio
async def test_method__with_stages():
    pipeline = Pipeline()

    stages: StageDefinitionCollection = (
        AddOneStage,  # stage class
        TimesTwoStage(),  # stage instance
    )

    new_pipeline = pipeline.with_stages(stages)

    assert new_pipeline is not pipeline  # ensures immutability
    assert len(new_pipeline.stages) == len(stages)
    for stage in stages:
        assert stage in new_pipeline.stages


@pytest.mark.asyncio
async def test_method__with_stages__override_current_instances():
    pipeline = Pipeline().pipe(AddOneStage())

    stages: StageDefinitionCollection = (
        TimesTwoStage,  # stage class
        TimesThreeStage(),  # stage instance
    )

    new_pipeline = pipeline.with_stages(stages)

    assert len(pipeline.get_stages_instances()) == 1
    assert len(new_pipeline.get_stages_instances()) == 2

    assert await pipeline.process(5) == 6
    assert await new_pipeline.process(5) == 30



from typing import Any, Optional, Protocol, Self

from ..support.act_as_factory.act_as_factory_interface import ActAsFactoryInterface
from ..types.callable_type import StageDefinition, StageDefinitionCollection
from ..types.types import T_in, T_out
from .pipeline import Pipeline as PipelineContract


class PipelineFactory(
    ActAsFactoryInterface[PipelineContract[T_in, T_out]],
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        stages: Optional[StageDefinitionCollection] = None,
        pipeline_class: Optional[type[PipelineContract[T_in, T_out]]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def add_stage(self, stage: StageDefinition) -> Self: ...

    def with_stages(self, stages: StageDefinitionCollection) -> Self: ...



from typing import Any

from ...pipeline import Pipeline
from ...types.types import T_in, T_out
from .interruptible_processor import CheckCallable, InterruptibleProcessor


class InterruptiblePipeline(Pipeline[T_in, T_out]):
    """Pipeline with conditional interruption."""

    def __init__(self, check: CheckCallable, *args: Any, **kwargs: Any):
        """Constructor.

        Parameters:
            check: Callable used to interrupt processing.

        Example:
            ```python
            # Interrupts when payload value exceeds 100
            pipeline = (
                InterruptiblePipeline[int](lambda payload: payload > 100)
                .pipe(lambda payload: payload + 2)
                .pipe(lambda payload: payload * 10)
                .pipe(lambda payload: payload * 10)
            )

            # Process payload - will stop if value exceeds 100
            assert await pipeline.process(5) == 70
            ```
        """
        processor = InterruptibleProcessor[T_in, T_out](check)

        super().__init__(processor=processor, *args, **kwargs)



from abc import abstractmethod
from typing import Any, Protocol, TypeVar

T_target = TypeVar("T_target", infer_variance=True)


class ActAsFactoryInterface(Protocol[T_target]):
    @abstractmethod
    def _definition(self) -> dict[str, Any]: ...

    def _get_target_class(self) -> type[T_target]: ...

    def make(self) -> Any: ...



"""Library's public types"""

# Re-exporting symbols
from _lib.types.callable_type import CallableCollection as CallableCollection
from _lib.types.callable_type import CallableType as CallableType
from _lib.types.callable_type import StageDefinition as StageDefinition
from _lib.types.callable_type import (
    StageDefinitionCollection as StageDefinitionCollection,
)
from _lib.types.types import T_in as T_in
from _lib.types.types import T_out as T_out

__all__ = (
    "CallableType",
    "CallableCollection",
    "StageDefinition",
    "StageDefinitionCollection",
    "T_in",
    "T_out",
)



from thecodecrate_pipeline import Processor
from thecodecrate_pipeline.types import CallableCollection


class StubProcessor(Processor[int]):
    async def process(
        self,
        payload: int,
        stages: CallableCollection,
    ) -> int:
        for stage in stages:
            payload = await self._call(callable=stage, payload=payload)

        return payload * 10



import subprocess


def test_ruff_compliance():
    """Test to ensure codebase is compliant with Ruff linting."""

    result = subprocess.run(
        ["uvx", "ruff", "check", "."],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    assert result.returncode == 0, f"Ruff found issues:\n{result.stdout}"


def test_zizmor_compliance():
    """Test to ensure codebase is compliant with Zizmor linting."""

    result = subprocess.run(
        ["uvx", "zizmor", "--min-severity", "low", ".github/workflows/"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    assert result.returncode == 0, f"Zizmor found issues:\n{result.stdout}"
